@@ -45,9 +45,13 @@
 // this many points.
 #define textPoolStep 4
 
-// Inter-character space width which will cause addChar to break up a
-// text string.
-#define defaultSpaceWidth 0.25
+// Inter-character space width which will cause addChar to start a new
+// word.
+#define minWordBreakSpace 0.1
+
+// Negative inter-character space width, i.e., overlap, which will
+// cause addChar to start a new word.
+#define minDupBreakOverlap 0.2
 
 // Max distance between baselines of two lines within a block, as a
 // fraction of the font size.
@@ -72,13 +76,17 @@
 
 // Minimum inter-word spacing, as a fraction of the font size.  (Only
 // used for raw ordering.)
-#define minWordSpacing 0.2
+#define minWordSpacing 0.15
 
 // Maximum inter-word spacing, as a fraction of the font size.
 #define maxWordSpacing 1.5
 
+// Maximum horizontal spacing which will allow a word to be pulled
+// into a block.
+#define minColSpacing1 0.3
+
 // Minimum spacing between columns, as a fraction of the font size.
-#define minColSpacing 1.0
+#define minColSpacing2 1.0
 
 // Maximum vertical spacing between blocks within a flow, as a
 // multiple of the font size.
@@ -213,9 +221,9 @@ TextWord::TextWord(GfxState *state, int rotA, double x0, double y0,
   } else {
     state->getFillRGB(&rgb);
   }
-  colorR = rgb.r;
-  colorG = rgb.g;
-  colorB = rgb.b;
+  colorR = colToDbl(rgb.r);
+  colorG = colToDbl(rgb.g);
+  colorB = colToDbl(rgb.b);
 #endif
 }
 
@@ -228,8 +236,8 @@ void TextWord::addChar(GfxState *state, double x, double y,
 		       double dx, double dy, Unicode u) {
   if (len == size) {
     size += 16;
-    text = (Unicode *)grealloc(text, size * sizeof(Unicode));
-    edge = (double *)grealloc(edge, (size + 1) * sizeof(double));
+    text = (Unicode *)greallocn(text, size, sizeof(Unicode));
+    edge = (double *)greallocn(edge, size + 1, sizeof(double));
   }
   text[len] = u;
   switch (rot) {
@@ -282,8 +290,8 @@ void TextWord::merge(TextWord *word) {
   }
   if (len + word->len > size) {
     size = len + word->len;
-    text = (Unicode *)grealloc(text, size * sizeof(Unicode));
-    edge = (double *)grealloc(edge, (size + 1) * sizeof(double));
+    text = (Unicode *)greallocn(text, size, sizeof(Unicode));
+    edge = (double *)greallocn(edge, size + 1, sizeof(double));
   }
   for (i = 0; i < word->len; ++i) {
     text[len + i] = word->text[i];
@@ -418,15 +426,15 @@ void TextPool::addWord(TextWord *word) {
   if (minBaseIdx > maxBaseIdx) {
     minBaseIdx = wordBaseIdx - 128;
     maxBaseIdx = wordBaseIdx + 128;
-    pool = (TextWord **)gmalloc((maxBaseIdx - minBaseIdx + 1) *
-				sizeof(TextWord *));
+    pool = (TextWord **)gmallocn(maxBaseIdx - minBaseIdx + 1,
+				 sizeof(TextWord *));
     for (baseIdx = minBaseIdx; baseIdx <= maxBaseIdx; ++baseIdx) {
       pool[baseIdx - minBaseIdx] = NULL;
     }
   } else if (wordBaseIdx < minBaseIdx) {
     newMinBaseIdx = wordBaseIdx - 128;
-    newPool = (TextWord **)gmalloc((maxBaseIdx - newMinBaseIdx + 1) *
-				   sizeof(TextWord *));
+    newPool = (TextWord **)gmallocn(maxBaseIdx - newMinBaseIdx + 1,
+				    sizeof(TextWord *));
     for (baseIdx = newMinBaseIdx; baseIdx < minBaseIdx; ++baseIdx) {
       newPool[baseIdx - newMinBaseIdx] = NULL;
     }
@@ -437,8 +445,8 @@ void TextPool::addWord(TextWord *word) {
     minBaseIdx = newMinBaseIdx;
   } else if (wordBaseIdx > maxBaseIdx) {
     newMaxBaseIdx = wordBaseIdx + 128;
-    pool = (TextWord **)grealloc(pool, (newMaxBaseIdx - minBaseIdx + 1) *
-				         sizeof(TextWord *));
+    pool = (TextWord **)greallocn(pool, newMaxBaseIdx - minBaseIdx + 1,
+				  sizeof(TextWord *));
     for (baseIdx = maxBaseIdx + 1; baseIdx <= newMaxBaseIdx; ++baseIdx) {
       pool[baseIdx - minBaseIdx] = NULL;
     }
@@ -660,8 +668,8 @@ void TextLine::coalesce(UnicodeMap *uMap) {
       ++len;
     }
   }
-  text = (Unicode *)gmalloc(len * sizeof(Unicode));
-  edge = (double *)gmalloc((len + 1) * sizeof(double));
+  text = (Unicode *)gmallocn(len, sizeof(Unicode));
+  edge = (double *)gmallocn(len + 1, sizeof(double));
   i = 0;
   for (word1 = words; word1; word1 = word1->next) {
     for (j = 0; j < word1->len; ++j) {
@@ -677,7 +685,7 @@ void TextLine::coalesce(UnicodeMap *uMap) {
   }
 
   // compute convertedLen and set up the col array
-  col = (int *)gmalloc((len + 1) * sizeof(int));
+  col = (int *)gmallocn(len + 1, sizeof(int));
   convertedLen = 0;
   for (i = 0; i < len; ++i) {
     col[i] = convertedLen;
@@ -861,22 +869,22 @@ int TextLineFrag::cmpYXPrimaryRot(const void *p1, const void *p2) {
   cmp = 0; // make gcc happy
   switch (frag1->line->blk->page->primaryRot) {
   case 0:
-    if ((cmp = frag1->yMin - frag2->yMin) == 0) {
+    if (fabs(cmp = frag1->yMin - frag2->yMin) < 0.01) {
       cmp = frag1->xMin - frag2->xMin;
     }
     break;
   case 1:
-    if ((cmp = frag2->xMax - frag1->xMax) == 0) {
+    if (fabs(cmp = frag2->xMax - frag1->xMax) < 0.01) {
       cmp = frag1->yMin - frag2->yMin;
     }
     break;
   case 2:
-    if ((cmp = frag2->yMin - frag1->yMin) == 0) {
+    if (fabs(cmp = frag2->yMin - frag1->yMin) < 0.01) {
       cmp = frag2->xMax - frag1->xMax;
     }
     break;
   case 3:
-    if ((cmp = frag1->xMax - frag2->xMax) == 0) {
+    if (fabs(cmp = frag1->xMax - frag2->xMax) < 0.01) {
       cmp = frag2->yMax - frag1->yMax;
     }
     break;
@@ -1182,7 +1190,7 @@ void TextBlock::coalesce(UnicodeMap *uMap) {
   }
 
   // sort lines into xy order for column assignment
-  lineArray = (TextLine **)gmalloc(nLines * sizeof(TextLine *));
+  lineArray = (TextLine **)gmallocn(nLines, sizeof(TextLine *));
   for (line = lines, i = 0; line; line = line->next, ++i) {
     lineArray[i] = line;
   }
@@ -1530,7 +1538,7 @@ TextWordList::TextWordList(TextPage *text, GBool physLayout) {
 	}
       }
     }
-    wordArray = (TextWord **)gmalloc(nWords * sizeof(TextWord *));
+    wordArray = (TextWord **)gmallocn(nWords, sizeof(TextWord *));
     i = 0;
     for (flow = text->flows; flow; flow = flow->next) {
       for (blk = flow->blocks; blk; blk = blk->next) {
@@ -1591,6 +1599,7 @@ TextPage::TextPage(GBool rawOrderA) {
   curFontSize = 0;
   nest = 0;
   nTinyChars = 0;
+  lastCharOverlap = gFalse;
   if (!rawOrder) {
     for (rot = 0; rot < 4; ++rot) {
       pools[rot] = new TextPool();
@@ -1627,6 +1636,12 @@ void TextPage::startPage(GfxState *state) {
   }
 }
 
+void TextPage::endPage() {
+  if (curWord) {
+    endWord();
+  }
+}
+
 void TextPage::clear() {
   int rot;
   TextFlow *flow;
@@ -1742,25 +1757,20 @@ void TextPage::updateFont(GfxState *state) {
 }
 
 void TextPage::beginWord(GfxState *state, double x0, double y0) {
-  double *txtm, *ctm, *fontm;
+  double *fontm;
   double m[4], m2[4];
   int rot;
 
   // This check is needed because Type 3 characters can contain
   // text-drawing operations (when TextPage is being used via
-  // XOutputDev rather than TextOutputDev).
+  // {X,Win}SplashOutputDev rather than TextOutputDev).
   if (curWord) {
     ++nest;
     return;
   }
 
   // compute the rotation
-  txtm = state->getTextMat();
-  ctm = state->getCTM();
-  m[0] = txtm[0] * ctm[0] + txtm[1] * ctm[2];
-  m[1] = txtm[0] * ctm[1] + txtm[1] * ctm[3];
-  m[2] = txtm[2] * ctm[0] + txtm[3] * ctm[2];
-  m[3] = txtm[2] * ctm[1] + txtm[3] * ctm[3];
+  state->getFontTransMat(&m[0], &m[1], &m[2], &m[3]);
   if (state->getFont()->getType() == fontType3) {
     fontm = state->getFont()->getFontMatrix();
     m2[0] = fontm[0] * m[0] + fontm[1] * m[2];
@@ -1783,20 +1793,16 @@ void TextPage::beginWord(GfxState *state, double x0, double y0) {
 
 void TextPage::addChar(GfxState *state, double x, double y,
 		       double dx, double dy,
-		       CharCode c, Unicode *u, int uLen) {
-  double x1, y1, w1, h1, dx2, dy2, sp;
-  int n, i;
-
-  // if the previous char was a space, addChar will have called
-  // endWord, so we need to start a new word
-  if (!curWord) {
-    beginWord(state, x, y);
-  }
+		       CharCode c, int nBytes, Unicode *u, int uLen) {
+  double x1, y1, w1, h1, dx2, dy2, base, sp, delta;
+  GBool overlap;
+  int i;
 
   // throw away chars that aren't inside the page bounds
   state->transform(x, y, &x1, &y1);
   if (x1 < 0 || x1 > pageWidth ||
       y1 < 0 || y1 > pageHeight) {
+    charPos += nBytes;
     return;
   }
 
@@ -1814,66 +1820,104 @@ void TextPage::addChar(GfxState *state, double x, double y,
   if (!globalParams->getTextKeepTinyChars() &&
       fabs(w1) < 3 && fabs(h1) < 3) {
     if (++nTinyChars > 50000) {
+      charPos += nBytes;
       return;
     }
   }
 
   // break words at space character
   if (uLen == 1 && u[0] == (Unicode)0x20) {
-    ++curWord->charLen;
-    ++charPos;
+    if (curWord) {
+      ++curWord->charLen;
+    }
+    charPos += nBytes;
     endWord();
     return;
   }
 
-  // large char spacing is sometimes used to move text around -- in
-  // this case, break text into individual chars and let the coalesce
-  // function deal with it later
-  n = curWord->len;
-  if (n > 0) {
+  // start a new word if:
+  // (1) this character doesn't fall in the right place relative to
+  //     the end of the previous word (this places upper and lower
+  //     constraints on the position deltas along both the primary
+  //     and secondary axes), or
+  // (2) this character overlaps the previous one (duplicated text), or
+  // (3) the previous character was an overlap (we want each duplicated
+  //     character to be in a word by itself at this stage)
+  if (curWord && curWord->len > 0) {
+    base = sp = delta = 0; // make gcc happy
     switch (curWord->rot) {
-    case 0: sp = x1 - curWord->xMax; break;
-    case 1: sp = y1 - curWord->yMax; break;
-    case 2: sp = curWord->xMin - x1; break;
-    case 3: sp = curWord->yMin - y1; break;
+    case 0:
+      base = y1;
+      sp = x1 - curWord->xMax;
+      delta = x1 - curWord->edge[curWord->len - 1];
+      break;
+    case 1:
+      base = x1;
+      sp = y1 - curWord->yMax;
+      delta = y1 - curWord->edge[curWord->len - 1];
+      break;
+    case 2:
+      base = y1;
+      sp = curWord->xMin - x1;
+      delta = curWord->edge[curWord->len - 1] - x1;
+      break;
+    case 3:
+      base = x1;
+      sp = curWord->yMin - y1;
+      delta = curWord->edge[curWord->len - 1] - y1;
+      break;
     }
-    if (sp > defaultSpaceWidth * curWord->fontSize) {
+    overlap = fabs(delta) < dupMaxPriDelta * curWord->fontSize &&
+              fabs(base - curWord->base) < dupMaxSecDelta * curWord->fontSize;
+    if (overlap || lastCharOverlap ||
+	sp < -minDupBreakOverlap * curWord->fontSize ||
+	sp > minWordBreakSpace * curWord->fontSize ||
+	fabs(base - curWord->base) > 0.5) {
       endWord();
+    }
+    lastCharOverlap = overlap;
+  } else {
+    lastCharOverlap = gFalse;
+  }
+
+  if (uLen != 0) {
+    // start a new word if needed
+    if (!curWord) {
       beginWord(state, x, y);
     }
-  }
 
-  // page rotation and/or transform matrices can cause text to be
-  // drawn in reverse order -- in this case, swap the begin/end
-  // coordinates and break text into individual chars
-  if ((curWord->rot == 0 && w1 < 0) ||
-      (curWord->rot == 1 && h1 < 0) ||
-      (curWord->rot == 2 && w1 > 0) ||
-      (curWord->rot == 3 && h1 > 0)) {
-    endWord();
-    beginWord(state, x + dx, y + dy);
-    x1 += w1;
-    y1 += h1;
-    w1 = -w1;
-    h1 = -h1;
-  }
+    // page rotation and/or transform matrices can cause text to be
+    // drawn in reverse order -- in this case, swap the begin/end
+    // coordinates and break text into individual chars
+    if ((curWord->rot == 0 && w1 < 0) ||
+	(curWord->rot == 1 && h1 < 0) ||
+	(curWord->rot == 2 && w1 > 0) ||
+	(curWord->rot == 3 && h1 > 0)) {
+      endWord();
+      beginWord(state, x + dx, y + dy);
+      x1 += w1;
+      y1 += h1;
+      w1 = -w1;
+      h1 = -h1;
+    }
 
-  // add the characters to the current word
-  if (uLen != 0) {
+    // add the characters to the current word
     w1 /= uLen;
     h1 /= uLen;
+    for (i = 0; i < uLen; ++i) {
+      curWord->addChar(state, x1 + i*w1, y1 + i*h1, w1, h1, u[i]);
+    }
   }
-  for (i = 0; i < uLen; ++i) {
-    curWord->addChar(state, x1 + i*w1, y1 + i*h1, w1, h1, u[i]);
+  if (curWord) {
+    curWord->charLen += nBytes;
   }
-  ++curWord->charLen;
-  ++charPos;
+  charPos += nBytes;
 }
 
 void TextPage::endWord() {
   // This check is needed because Type 3 characters can contain
   // text-drawing operations (when TextPage is being used via
-  // XOutputDev rather than TextOutputDev).
+  // {X,Win}SplashOutputDev rather than TextOutputDev).
   if (nest > 0) {
     --nest;
     return;
@@ -1915,7 +1959,7 @@ void TextPage::coalesce(GBool physLayout) {
   TextFlow *flow, *lastFlow;
   int rot, poolMinBaseIdx, baseIdx, startBaseIdx;
   double minBase, maxBase, newMinBase, newMaxBase;
-  double fontSize, colSpace, lineSpace, intraLineSpace, blkSpace;
+  double fontSize, colSpace1, colSpace2, lineSpace, intraLineSpace, blkSpace;
   GBool found;
   int count[4];
   int lrCount;
@@ -1941,9 +1985,9 @@ void TextPage::coalesce(GBool physLayout) {
     pool = pools[rot];
     for (baseIdx = pool->minBaseIdx; baseIdx <= pool->maxBaseIdx; ++baseIdx) {
       for (word0 = pool->getPool(baseIdx); word0; word0 = word0->next) {
-	printf("    word: x=%.2f..%.2f y=%.2f..%.2f base=%.2f fontSize=%.2f '",
+	printf("    word: x=%.2f..%.2f y=%.2f..%.2f base=%.2f fontSize=%.2f rot=%d '",
 	       word0->xMin, word0->xMax, word0->yMin, word0->yMax,
-	       word0->base, word0->fontSize);
+	       word0->base, word0->fontSize, rot*90);
 	for (i = 0; i < word0->len; ++i) {
 	  fputc(word0->text[i] & 0xff, stdout);
 	}
@@ -2000,7 +2044,8 @@ void TextPage::coalesce(GBool physLayout) {
 
       fontSize = word0->fontSize;
       minBase = maxBase = word0->base;
-      colSpace = minColSpacing * fontSize;
+      colSpace1 = minColSpacing1 * fontSize;
+      colSpace2 = minColSpacing2 * fontSize;
       lineSpace = maxLineSpacingDelta * fontSize;
       intraLineSpace = maxIntraLineDelta * fontSize;
 
@@ -2089,8 +2134,10 @@ void TextPage::coalesce(GBool physLayout) {
 	    if (word1->base >= minBase - intraLineSpace &&
 		word1->base <= maxBase + intraLineSpace &&
 		((rot == 0 || rot == 2)
-		 ? (word1->xMin < blk->xMax && word1->xMax > blk->xMin)
-		 : (word1->yMin < blk->yMax && word1->yMax > blk->yMin)) &&
+		 ? (word1->xMin < blk->xMax + colSpace1 &&
+		    word1->xMax > blk->xMin - colSpace1)
+		 : (word1->yMin < blk->yMax + colSpace1 &&
+		    word1->yMax > blk->yMin - colSpace1)) &&
 		fabs(word1->fontSize - fontSize) <
 		  maxBlockFontSizeDelta2 * fontSize) {
 	      word2 = word1;
@@ -2129,9 +2176,9 @@ void TextPage::coalesce(GBool physLayout) {
 		word1->base <= maxBase + intraLineSpace &&
 		((rot == 0 || rot == 2)
 		 ? (word1->xMax <= blk->xMin &&
-		    word1->xMax > blk->xMin - colSpace)
+		    word1->xMax > blk->xMin - colSpace2)
 		 : (word1->yMax <= blk->yMin &&
-		    word1->yMax > blk->yMin - colSpace)) &&
+		    word1->yMax > blk->yMin - colSpace2)) &&
 		fabs(word1->fontSize - fontSize) <
 		  maxBlockFontSizeDelta3 * fontSize) {
 	      ++n;
@@ -2151,9 +2198,9 @@ void TextPage::coalesce(GBool physLayout) {
 		  word1->base <= maxBase + intraLineSpace &&
 		  ((rot == 0 || rot == 2)
 		   ? (word1->xMax <= blk->xMin &&
-		      word1->xMax > blk->xMin - colSpace)
+		      word1->xMax > blk->xMin - colSpace2)
 		   : (word1->yMax <= blk->yMin &&
-		      word1->yMax > blk->yMin - colSpace)) &&
+		      word1->yMax > blk->yMin - colSpace2)) &&
 		  fabs(word1->fontSize - fontSize) <
 		    maxBlockFontSizeDelta3 * fontSize) {
 		word2 = word1;
@@ -2193,9 +2240,9 @@ void TextPage::coalesce(GBool physLayout) {
 		word1->base <= maxBase + intraLineSpace &&
 		((rot == 0 || rot == 2)
 		 ? (word1->xMin >= blk->xMax &&
-		    word1->xMin < blk->xMax + colSpace)
+		    word1->xMin < blk->xMax + colSpace2)
 		 : (word1->yMin >= blk->yMax &&
-		    word1->yMin < blk->yMax + colSpace)) &&
+		    word1->yMin < blk->yMax + colSpace2)) &&
 		fabs(word1->fontSize - fontSize) <
 		  maxBlockFontSizeDelta3 * fontSize) {
 	      ++n;
@@ -2215,9 +2262,9 @@ void TextPage::coalesce(GBool physLayout) {
 		  word1->base <= maxBase + intraLineSpace &&
 		  ((rot == 0 || rot == 2)
 		   ? (word1->xMin >= blk->xMax &&
-		      word1->xMin < blk->xMax + colSpace)
+		      word1->xMin < blk->xMax + colSpace2)
 		   : (word1->yMin >= blk->yMax &&
-		      word1->yMin < blk->yMax + colSpace)) &&
+		      word1->yMin < blk->yMax + colSpace2)) &&
 		  fabs(word1->fontSize - fontSize) <
 		    maxBlockFontSizeDelta3 * fontSize) {
 		word2 = word1;
@@ -2322,7 +2369,7 @@ void TextPage::coalesce(GBool physLayout) {
   //----- column assignment
 
   // sort blocks into xy order for column assignment
-  blocks = (TextBlock **)gmalloc(nBlocks * sizeof(TextBlock *));
+  blocks = (TextBlock **)gmallocn(nBlocks, sizeof(TextBlock *));
   for (blk = blkList, i = 0; blk; blk = blk->next, ++i) {
     blocks[i] = blk;
   }
@@ -2339,6 +2386,8 @@ void TextPage::coalesce(GBool physLayout) {
       case 0:
 	if (blk0->xMin > blk1->xMax) {
 	  col2 = blk1->col + blk1->nColumns + 3;
+	} else if (blk1->xMax == blk1->xMin) {
+	  col2 = blk1->col;
 	} else {
 	  col2 = blk1->col + (int)(((blk0->xMin - blk1->xMin) /
 				    (blk1->xMax - blk1->xMin)) *
@@ -2348,6 +2397,8 @@ void TextPage::coalesce(GBool physLayout) {
       case 1:
 	if (blk0->yMin > blk1->yMax) {
 	  col2 = blk1->col + blk1->nColumns + 3;
+	} else if (blk1->yMax == blk1->yMin) {
+	  col2 = blk1->col;
 	} else {
 	  col2 = blk1->col + (int)(((blk0->yMin - blk1->yMin) /
 				    (blk1->yMax - blk1->yMin)) *
@@ -2357,6 +2408,8 @@ void TextPage::coalesce(GBool physLayout) {
       case 2:
 	if (blk0->xMax < blk1->xMin) {
 	  col2 = blk1->col + blk1->nColumns + 3;
+	} else if (blk1->xMin == blk1->xMax) {
+	  col2 = blk1->col;
 	} else {
 	  col2 = blk1->col + (int)(((blk0->xMax - blk1->xMax) /
 				    (blk1->xMin - blk1->xMax)) *
@@ -2366,6 +2419,8 @@ void TextPage::coalesce(GBool physLayout) {
       case 3:
 	if (blk0->yMax < blk1->yMin) {
 	  col2 = blk1->col + blk1->nColumns + 3;
+	} else if (blk1->yMin == blk1->yMax) {
+	  col2 = blk1->col;
 	} else {
 	  col2 = blk1->col + (int)(((blk0->yMax - blk1->yMax) /
 				    (blk1->yMin - blk1->yMax)) *
@@ -2449,7 +2504,7 @@ void TextPage::coalesce(GBool physLayout) {
   // build the flows
   //~ this needs to be adjusted for writing mode (vertical text)
   //~ this also needs to account for right-to-left column ordering
-  blkArray = (TextBlock **)gmalloc(nBlocks * sizeof(TextBlock *));
+  blkArray = (TextBlock **)gmallocn(nBlocks, sizeof(TextBlock *));
   memcpy(blkArray, blocks, nBlocks * sizeof(TextBlock *));
   flows = lastFlow = NULL;
   firstBlkIdx = 0;
@@ -2570,13 +2625,14 @@ void TextPage::coalesce(GBool physLayout) {
 GBool TextPage::findText(Unicode *s, int len,
 			 GBool startAtTop, GBool stopAtBottom,
 			 GBool startAtLast, GBool stopAtLast,
+			 GBool caseSensitive, GBool backward,
 			 double *xMin, double *yMin,
 			 double *xMax, double *yMax) {
   TextBlock *blk;
   TextLine *line;
+  Unicode *s2, *txt;
   Unicode *p;
-  Unicode u1, u2;
-  int m, i, j, k;
+  int txtSize, m, i, j, k;
   double xStart, yStart, xStop, yStop;
   double xMin0, yMin0, xMax0, yMax0;
   double xMin1, yMin1, xMax1, yMax1;
@@ -2588,6 +2644,19 @@ GBool TextPage::findText(Unicode *s, int len,
     return gFalse;
   }
 
+  // convert the search string to uppercase
+  if (!caseSensitive) {
+    s2 = (Unicode *)gmallocn(len, sizeof(Unicode));
+    for (i = 0; i < len; ++i) {
+      s2[i] = unicodeToUpper(s[i]);
+    }
+  } else {
+    s2 = s;
+  }
+
+  txt = NULL;
+  txtSize = 0;
+
   xStart = yStart = xStop = yStop = 0;
   if (startAtLast && haveLastFind) {
     xStart = lastFindXMin;
@@ -2608,51 +2677,57 @@ GBool TextPage::findText(Unicode *s, int len,
   xMin0 = xMax0 = yMin0 = yMax0 = 0; // make gcc happy
   xMin1 = xMax1 = yMin1 = yMax1 = 0; // make gcc happy
 
-  for (i = 0; i < nBlocks; ++i) {
+  for (i = backward ? nBlocks - 1 : 0;
+       backward ? i >= 0 : i < nBlocks;
+       i += backward ? -1 : 1) {
     blk = blocks[i];
 
     // check: is the block above the top limit?
-    if (!startAtTop && blk->yMax < yStart) {
+    if (!startAtTop && (backward ? blk->yMin > yStart : blk->yMax < yStart)) {
       continue;
     }
 
     // check: is the block below the bottom limit?
-    if (!stopAtBottom && blk->yMin > yStop) {
+    if (!stopAtBottom && (backward ? blk->yMax < yStop : blk->yMin > yStop)) {
       break;
     }
 
     for (line = blk->lines; line; line = line->next) {
 
       // check: is the line above the top limit?
-      if (!startAtTop && line->yMin < yStart) {
+      if (!startAtTop &&
+	  (backward ? line->yMin > yStart : line->yMin < yStart)) {
 	continue;
       }
 
       // check: is the line below the bottom limit?
-      if (!stopAtBottom && line->yMin > yStop) {
+      if (!stopAtBottom &&
+	  (backward ? line->yMin < yStop : line->yMin > yStop)) {
 	continue;
       }
 
-      // search each position in this line
+      // convert the line to uppercase
       m = line->len;
-      for (j = 0, p = line->text; j <= m - len; ++j, ++p) {
+      if (!caseSensitive) {
+	if (m > txtSize) {
+	  txt = (Unicode *)greallocn(txt, m, sizeof(Unicode));
+	  txtSize = m;
+	}
+	for (k = 0; k < m; ++k) {
+	  txt[k] = unicodeToUpper(line->text[k]);
+	}
+      } else {
+	txt = line->text;
+      }
+
+      // search each position in this line
+      j = backward ? m - len : 0;
+      p = txt + j;
+      while (backward ? j >= 0 : j <= m - len) {
 
 	// compare the strings
 	for (k = 0; k < len; ++k) {
-#if 1 //~ this lowercases Latin A-Z only -- this will eventually be
-      //~ extended to handle other character sets
-	  if (p[k] >= 0x41 && p[k] <= 0x5a) {
-	    u1 = p[k] + 0x20;
-	  } else {
-	    u1 = p[k];
-	  }
-	  if (s[k] >= 0x41 && s[k] <= 0x5a) {
-	    u2 = s[k] + 0x20;
-	  } else {
-	    u2 = s[k];
-	  }
-#endif
-	  if (u1 != u2) {
+	  if (p[k] != s2[k]) {
 	    break;
 	  }
 	}
@@ -2685,23 +2760,52 @@ GBool TextPage::findText(Unicode *s, int len,
 	    yMax1 = line->edge[j];
 	    break;
 	  }
-	  if ((startAtTop ||
-	       yMin1 > yStart || (yMin1 == yStart && xMin1 > xStart)) &&
-	      (stopAtBottom ||
-	       yMin1 < yStop || (yMin1 == yStop && xMin1 < yStop))) {
-	    if (!found || yMin1 < yMin0 || (yMin1 == yMin0 && xMin1 < xMin0)) {
-	      xMin0 = xMin1;
-	      xMax0 = xMax1;
-	      yMin0 = yMin1;
-	      yMax0 = yMax1;
-	      found = gTrue;
+	  if (backward) {
+	    if ((startAtTop ||
+		 yMin1 < yStart || (yMin1 == yStart && xMin1 < xStart)) &&
+		(stopAtBottom ||
+		 yMin1 > yStop || (yMin1 == yStop && xMin1 > xStop))) {
+	      if (!found ||
+		  yMin1 > yMin0 || (yMin1 == yMin0 && xMin1 > xMin0)) {
+		xMin0 = xMin1;
+		xMax0 = xMax1;
+		yMin0 = yMin1;
+		yMax0 = yMax1;
+		found = gTrue;
+	      }
+	    }
+	  } else {
+	    if ((startAtTop ||
+		 yMin1 > yStart || (yMin1 == yStart && xMin1 > xStart)) &&
+		(stopAtBottom ||
+		 yMin1 < yStop || (yMin1 == yStop && xMin1 < xStop))) {
+	      if (!found ||
+		  yMin1 < yMin0 || (yMin1 == yMin0 && xMin1 < xMin0)) {
+		xMin0 = xMin1;
+		xMax0 = xMax1;
+		yMin0 = yMin1;
+		yMax0 = yMax1;
+		found = gTrue;
+	      }
 	    }
 	  }
 	}
+	if (backward) {
+	  --j;
+	  --p;
+	} else {
+	  ++j;
+	  ++p;
+	}
       }
     }
   }
 
+  if (!caseSensitive) {
+    gfree(s2);
+    gfree(txt);
+  }
+
   if (found) {
     *xMin = xMin0;
     *xMax = xMax0;
@@ -2763,7 +2867,7 @@ GString *TextPage::getText(double xMin, double yMin,
 
   // collect the line fragments that are in the rectangle
   fragsSize = 256;
-  frags = (TextLineFrag *)gmalloc(fragsSize * sizeof(TextLineFrag));
+  frags = (TextLineFrag *)gmallocn(fragsSize, sizeof(TextLineFrag));
   nFrags = 0;
   lastRot = -1;
   oneRot = gTrue;
@@ -2865,7 +2969,7 @@ GString *TextPage::getText(double xMin, double yMin,
 	    if (nFrags == fragsSize) {
 	      fragsSize *= 2;
 	      frags = (TextLineFrag *)
-		          grealloc(frags, fragsSize * sizeof(TextLineFrag));
+		          greallocn(frags, fragsSize, sizeof(TextLineFrag));
 	    }
 	    frags[nFrags].init(line, idx0, idx1 - idx0 + 1);
 	    ++nFrags;
@@ -3083,15 +3187,15 @@ void TextPage::dump(void *outputStream, TextOutputFunc outputFunc,
 
     // collect the line fragments for the page and sort them
     fragsSize = 256;
-    frags = (TextLineFrag *)gmalloc(fragsSize * sizeof(TextLineFrag));
+    frags = (TextLineFrag *)gmallocn(fragsSize, sizeof(TextLineFrag));
     nFrags = 0;
     for (i = 0; i < nBlocks; ++i) {
       blk = blocks[i];
       for (line = blk->lines; line; line = line->next) {
 	if (nFrags == fragsSize) {
 	  fragsSize *= 2;
-	  frags = (TextLineFrag *)grealloc(frags,
-					   fragsSize * sizeof(TextLineFrag));
+	  frags = (TextLineFrag *)greallocn(frags,
+					    fragsSize, sizeof(TextLineFrag));
 	}
 	frags[nFrags].init(line, 0, line->len);
 	frags[nFrags].computeCoords(gTrue);
@@ -3100,6 +3204,20 @@ void TextPage::dump(void *outputStream, TextOutputFunc outputFunc,
     }
     qsort(frags, nFrags, sizeof(TextLineFrag), &TextLineFrag::cmpYXPrimaryRot);
 
+#if 0 // for debugging
+    printf("*** line fragments ***\n");
+    for (i = 0; i < nFrags; ++i) {
+      frag = &frags[i];
+      printf("frag: x=%.2f..%.2f y=%.2f..%.2f base=%.2f '",
+	     frag->xMin, frag->xMax, frag->yMin, frag->yMax, frag->base);
+      for (n = 0; n < frag->len; ++n) {
+	fputc(frag->line->text[frag->start + n] & 0xff, stdout);
+      }
+      printf("'\n");
+    }
+    printf("\n");
+#endif
+
     // generate output
     col = 0;
     for (i = 0; i < nFrags; ++i) {
@@ -3178,7 +3296,6 @@ void TextPage::dump(void *outputStream, TextOutputFunc outputFunc,
   // end of page
   if (pageBreaks) {
     (*outputFunc)(outputStream, eop, eopLen);
-    (*outputFunc)(outputStream, eol, eolLen);
   }
 
   uMap->decRefCnt();
@@ -3435,6 +3552,7 @@ void TextOutputDev::startPage(int pageNum, GfxState *state) {
 }
 
 void TextOutputDev::endPage() {
+  text->endPage();
   text->coalesce(physLayout);
   if (outputStream) {
     text->dump(outputStream, outputFunc, physLayout);
@@ -3446,27 +3564,27 @@ void TextOutputDev::updateFont(GfxState *state) {
 }
 
 void TextOutputDev::beginString(GfxState *state, GString *s) {
-  text->beginWord(state, state->getCurX(), state->getCurY());
 }
 
 void TextOutputDev::endString(GfxState *state) {
-  text->endWord();
 }
 
 void TextOutputDev::drawChar(GfxState *state, double x, double y,
 			     double dx, double dy,
 			     double originX, double originY,
-			     CharCode c, Unicode *u, int uLen) {
-  text->addChar(state, x, y, dx, dy, c, u, uLen);
+			     CharCode c, int nBytes, Unicode *u, int uLen) {
+  text->addChar(state, x, y, dx, dy, c, nBytes, u, uLen);
 }
 
 GBool TextOutputDev::findText(Unicode *s, int len,
 			      GBool startAtTop, GBool stopAtBottom,
 			      GBool startAtLast, GBool stopAtLast,
+			      GBool caseSensitive, GBool backward,
 			      double *xMin, double *yMin,
 			      double *xMax, double *yMax) {
   return text->findText(s, len, startAtTop, stopAtBottom,
-			startAtLast, stopAtLast, xMin, yMin, xMax, yMax);
+			startAtLast, stopAtLast, caseSensitive, backward,
+			xMin, yMin, xMax, yMax);
 }
 
 GString *TextOutputDev::getText(double xMin, double yMin,
@@ -3485,3 +3603,11 @@ TextWordList *TextOutputDev::makeWordList() {
   return text->makeWordList(physLayout);
 }
 #endif
+
+TextPage *TextOutputDev::takeText() {
+  TextPage *ret;
+
+  ret = text;
+  text = new TextPage(rawOrder);
+  return ret;
+}
