@@ -26,6 +26,15 @@ class GfxFont;
 class GfxState;
 class UnicodeMap;
 
+class TextWord;
+class TextPool;
+class TextLine;
+class TextLineFrag;
+class TextBlock;
+class TextFlow;
+class TextWordList;
+class TextPage;
+
 //------------------------------------------------------------------------
 
 typedef void (*TextOutputFunc)(void *stream, char *text, int len);
@@ -93,6 +102,8 @@ public:
     { *r = colorR; *g = colorG; *b = colorB; }
   void getBBox(double *xMinA, double *yMinA, double *xMaxA, double *yMaxA)
     { *xMinA = xMin; *yMinA = yMin; *xMaxA = xMax; *yMaxA = yMax; }
+  double getFontSize() { return fontSize; }
+  int getRotation() { return rot; }
   int getCharPos() { return charPos; }
   int getCharLen() { return charLen; }
 #endif
@@ -354,6 +365,9 @@ public:
   // Start a new page.
   void startPage(GfxState *state);
 
+  // End the current page.
+  void endPage();
+
   // Update the current font.
   void updateFont(GfxState *state);
 
@@ -363,7 +377,7 @@ public:
   // Add a character to the current word.
   void addChar(GfxState *state, double x, double y,
 	       double dx, double dy,
-	       CharCode c, Unicode *u, int uLen);
+	       CharCode c, int nBytes, Unicode *u, int uLen);
 
   // End the current word, sorting it into the list of words.
   void endWord();
@@ -384,6 +398,7 @@ public:
   GBool findText(Unicode *s, int len,
 		 GBool startAtTop, GBool stopAtBottom,
 		 GBool startAtLast, GBool stopAtLast,
+		 GBool caseSensitive, GBool backward,
 		 double *xMin, double *yMin,
 		 double *xMax, double *yMax);
 
@@ -426,6 +441,8 @@ private:
   double curFontSize;		// current font size
   int nest;			// current nesting level (for Type 3 fonts)
   int nTinyChars;		// number of "tiny" chars seen so far
+  GBool lastCharOverlap;	// set if the last added char overlapped the
+				//   previous char
 
   TextPool *pools[4];		// a "pool" of TextWords for each rotation
   TextFlow *flows;		// linked list of flows
@@ -513,7 +530,7 @@ public:
   virtual void drawChar(GfxState *state, double x, double y,
 			double dx, double dy,
 			double originX, double originY,
-			CharCode c, Unicode *u, int uLen);
+			CharCode c, int nBytes, Unicode *u, int uLen);
 
   //----- special access
 
@@ -527,6 +544,7 @@ public:
   GBool findText(Unicode *s, int len,
 		 GBool startAtTop, GBool stopAtBottom,
 		 GBool startAtLast, GBool stopAtLast,
+		 GBool caseSensitive, GBool backward,
 		 double *xMin, double *yMin,
 		 double *xMax, double *yMax);
 
@@ -549,6 +567,10 @@ public:
   TextWordList *makeWordList();
 #endif
 
+  // Returns the TextPage object for the last rasterized page,
+  // transferring ownership to the caller.
+  TextPage *takeText();
+
 private:
 
   TextOutputFunc outputFunc;	// output function
