@@ -144,7 +144,8 @@ int main(int argc, char *argv[]) {
     if ((resDict = page->getResourceDict())) {
       scanFonts(resDict, doc);
     }
-    annots = new Annots(doc->getXRef(), page->getAnnots(&obj1));
+    annots = new Annots(doc->getXRef(), doc->getCatalog(),
+			page->getAnnots(&obj1));
     obj1.free();
     for (i = 0; i < annots->getNumAnnots(); ++i) {
       if (annots->getAnnot(i)->getAppearance(&obj1)->isStream()) {
@@ -197,8 +198,9 @@ static void scanFonts(Dict *resDict, PDFDoc *doc) {
   }
   if (gfxFontDict) {
     for (i = 0; i < gfxFontDict->getNumFonts(); ++i) {
-      font = gfxFontDict->getFont(i);
-      scanFont(font, doc);
+      if ((font = gfxFontDict->getFont(i))) {
+	scanFont(font, doc);
+      }
     }
     delete gfxFontDict;
   }
@@ -225,9 +227,9 @@ static void scanFonts(Dict *resDict, PDFDoc *doc) {
 
 static void scanFont(GfxFont *font, PDFDoc *doc) {
   Ref fontRef, embRef;
-  Object fontObj, nameObj, toUnicodeObj;
+  Object fontObj, toUnicodeObj;
   GString *name;
-  GBool subset, hasToUnicode;
+  GBool emb, subset, hasToUnicode;
   int i;
 
   fontRef = *font->getID();
@@ -242,6 +244,13 @@ static void scanFont(GfxFont *font, PDFDoc *doc) {
   // font name
   name = font->getOrigName();
 
+  // check for an embedded font
+  if (font->getType() == fontType3) {
+    emb = gTrue;
+  } else {
+    emb = font->getEmbeddedFontID(&embRef);
+  }
+
   // look for a ToUnicode map
   hasToUnicode = gFalse;
   if (doc->getXRef()->fetch(fontRef.num, fontRef.gen, &fontObj)->isDict()) {
@@ -266,7 +275,7 @@ static void scanFont(GfxFont *font, PDFDoc *doc) {
   printf("%-36s %-12s %-3s %-3s %-3s",
 	 name ? name->getCString() : "[none]",
 	 fontTypeNames[font->getType()],
-	 font->getEmbeddedFontID(&embRef) ? "yes" : "no",
+	 emb ? "yes" : "no",
 	 subset ? "yes" : "no",
 	 hasToUnicode ? "yes" : "no");
   if (fontRef.gen >= 100000) {
@@ -278,7 +287,7 @@ static void scanFont(GfxFont *font, PDFDoc *doc) {
   // add this font to the list
   if (fontsLen == fontsSize) {
     fontsSize += 32;
-    fonts = (Ref *)grealloc(fonts, fontsSize * sizeof(Ref));
+    fonts = (Ref *)greallocn(fonts, fontsSize, sizeof(Ref));
   }
   fonts[fontsLen++] = *font->getID();
 }
